from pydantic_settings import SettingsError
import pytest


def test_missing_dir_ok(settings_model, monkeypatch, tmp_path):
    monkeypatch.setenv('DB__USER', 'user')
    Settings = settings_model(
        model_config=dict(
            env_nested_delimiter='__',
            secrets_dir=tmp_path / 'non-existing',
            secrets_dir_missing='ok',
        ),
    )
    conf = Settings()
    assert conf.app_key is None
    assert conf.db.password is None


def test_missing_dir_warn(settings_model, monkeypatch, tmp_path):
    monkeypatch.setenv('DB__USER', 'user')
    Settings = settings_model(
        model_config=dict(
            env_nested_delimiter='__',
            secrets_dir=tmp_path / 'non-existing',
            # warn is default
        ),
    )
    with pytest.warns(UserWarning, match='does not exist'):
        conf = Settings()
    assert conf.app_key is None
    assert conf.db.password is None


def test_missing_dir_error(settings_model, monkeypatch, tmp_path):
    monkeypatch.setenv('DB__USER', 'user')
    Settings = settings_model(
        model_config=dict(
            env_nested_delimiter='__',
            secrets_dir=tmp_path / 'non-existing',
            secrets_dir_missing='error',
        ),
    )
    with pytest.raises(SettingsError, match='does not exist'):
        Settings()


def test_missing_dir_invalid(settings_model, monkeypatch, tmp_path):
    monkeypatch.setenv('DB__USER', 'user')
    Settings = settings_model(
        model_config=dict(
            env_nested_delimiter='__',
            secrets_dir=tmp_path / 'non-existing',
            secrets_dir_missing='whatever',  # invalid value
        ),
    )
    with pytest.raises(SettingsError, match='invalid secrets_dir_missing value'):
        Settings()


def test_secrets_not_dir(settings_model, monkeypatch, secrets_dir):
    monkeypatch.setenv('DB__USER', 'user')
    secrets_dir.add_files(
        ('secrets_notdir', ''),
    )
    Settings = settings_model(
        model_config=dict(
            env_nested_delimiter='__',
            secrets_dir=secrets_dir / 'secrets_notdir',  # file
        ),
    )
    with pytest.raises(SettingsError, match='must reference a directory'):
        Settings()


def test_secrets_dir_size(settings_model, monkeypatch, secrets_dir):
    SIZE = 10
    monkeypatch.setenv('DB__USER', 'user')
    secrets_dir.add_files(
        ('large_file', ' ' * SIZE),
    )
    Settings = settings_model(
        model_config=dict(
            env_nested_delimiter='__',
            secrets_dir=secrets_dir,
            secrets_dir_max_size=SIZE - 1,
        ),
    )
    with pytest.raises(SettingsError, match='secrets_dir size'):
        Settings()



from pydantic_file_secrets import FileSecretsSettingsSource
from pydantic_settings import BaseSettings, SettingsConfigDict


def test_repr(secrets_dir):
    class Settings(BaseSettings):
        model_config = SettingsConfigDict(
            secrets_dir=secrets_dir,
        )

    src = FileSecretsSettingsSource(Settings)
    assert f'{src!r}'.startswith(f'{src.__class__.__name__}(')



def test_prefix_plain(settings_model, monkeypatch, secrets_dir):
    monkeypatch.setenv('DB__USER', 'user')
    secrets_dir.add_files(
        ('prefix_app_key', 'secret1'),
        ('prefix_db___password', 'secret2'),
    )
    Settings = settings_model(
        model_config=dict(
            env_nested_delimiter='__',
            secrets_dir=secrets_dir,
            secrets_nested_delimiter='___',
            secrets_prefix='prefix_',
        ),
    )
    conf = Settings()
    assert conf.app_key == 'secret1'
    assert conf.db.password == 'secret2'  # noqa: S105


def test_prefix_with_subdir(settings_model, monkeypatch, secrets_dir):
    monkeypatch.setenv('DB__USER', 'user')
    secrets_dir.add_files(
        ('prefix_app_key', 'secret1'),
        ('prefix_db/password', 'secret2'),
    )
    Settings = settings_model(
        model_config=dict(
            env_nested_delimiter='__',
            secrets_dir=secrets_dir,
            secrets_nested_subdir=True,
            secrets_prefix='prefix_',
        ),
    )
    conf = Settings()
    assert conf.app_key == 'secret1'
    assert conf.db.password == 'secret2'  # noqa: S105


def test_prefix_multiple_subdirs(settings_model, monkeypatch, secrets_dir):
    monkeypatch.setenv('DB__USER', 'user')
    secrets_dir.add_files(
        ('dir1/dir2/app_key', 'secret1'),
        ('dir1/dir2/db/password', 'secret2'),
    )
    Settings = settings_model(
        model_config=dict(
            env_nested_delimiter='__',
            secrets_dir=secrets_dir,
            secrets_nested_subdir=True,
            secrets_prefix='dir1/dir2/',
        ),
    )
    conf = Settings()
    assert conf.app_key == 'secret1'
    assert conf.db.password == 'secret2'  # noqa: S105



"""
These tests show that the interface of original SecretsSettingsSource
declares features that are not working.
"""

from enum import StrEnum

from pydantic_settings import BaseSettings, SettingsConfigDict


def test_not_working_env_ignore_empty(secrets_dir):
    class TestEnum(StrEnum):
        TEST = 'test'

    class Settings(BaseSettings):
        key_empty: str
        key_none: str
        key_enum: TestEnum

        model_config = SettingsConfigDict(
            secrets_dir=secrets_dir,
            # options below don't work
            env_ignore_empty=True,
            env_parse_none_str='null',
            env_parse_enums=False,
        )

    secrets_dir.add_files(
        ('key_empty', ''),
        ('key_none', 'null'),
        ('key_enum', 'test'),
    )

    conf = Settings()
    assert conf.key_empty == ''  # should be None if working
    assert conf.key_none == 'null'  # should be Null if working
    assert isinstance(conf.key_enum, TestEnum)  # should be True if working



from pathlib import Path

from pydantic import BaseModel
from pydantic_settings import BaseSettings, SettingsConfigDict
from pytest import fixture

from pydantic_file_secrets import FileSecretsSettingsSource


# Settings


class DbSettings(BaseModel):
    user: str
    password: str | None = None


class Settings(BaseSettings):
    db: DbSettings
    app_key: str | None = None


class SettingsMaker:
    def __call__(
        self,
        model_config: SettingsConfigDict | dict | None,
    ) -> type[Settings]:
        class TestSettings(Settings):
            @classmethod
            def settings_customise_sources(
                cls,
                settings_cls,
                init_settings,
                env_settings,
                dotenv_settings,
                file_secret_settings,
            ) -> tuple:
                return (
                    env_settings,
                    init_settings,
                    FileSecretsSettingsSource(settings_cls),
                )

        TestSettings.model_config = model_config or {}
        return TestSettings


@fixture()
def settings_model() -> SettingsMaker:
    return SettingsMaker()


# secrets_dir


class SecretsDir(Path):
    def add_files(self, *files: tuple[str, str]) -> None:
        for path, content in files:
            f = self / path
            f.parent.mkdir(parents=True, exist_ok=True)  # allow child dirs in path
            f.write_text(content)


@fixture()
def secrets_dir(tmp_path) -> SecretsDir:
    yield SecretsDir(tmp_path)



from pydantic_settings import SettingsError
import pytest


def test_subdir(settings_model, monkeypatch, secrets_dir):
    monkeypatch.setenv('DB__USER', 'user')
    secrets_dir.add_files(
        ('app_key', 'secret1'),
        ('db/password', 'secret2'),  # file in subdir
    )
    Settings = settings_model(
        model_config=dict(
            secrets_dir=secrets_dir,
            env_nested_delimiter='__',
            secrets_nested_subdir=True,
        ),
    )
    conf = Settings()
    assert conf.app_key == 'secret1'
    assert conf.db.password == 'secret2'  # noqa: S105


def test_invalid_options(settings_model, monkeypatch, secrets_dir):
    monkeypatch.setenv('DB__USER', 'user')
    Settings = settings_model(
        model_config=dict(
            secrets_dir=secrets_dir,
            env_nested_delimiter='__',
            secrets_nested_subdir=True,
            secrets_nested_delimiter='__',
        ),
    )
    with pytest.raises(SettingsError, match='mutually exclusive'):
        Settings()



import os
from pathlib import Path
from typing import Any, Literal
import warnings

from pydantic_settings import BaseSettings, EnvSettingsSource, SettingsError
from pydantic_settings.sources import parse_env_vars
from pydantic_settings.utils import path_type_label

from .__version__ import __version__


__all__ = ['__version__', 'FileSecretsSettingsSource']


type SecretsDirMissing = Literal['ok', 'warn', 'error']


class FileSecretsSettingsSource(EnvSettingsSource):
    def __init__(
        self,
        settings_cls: type[BaseSettings],
        secrets_dir: str | Path | None = None,
        secrets_dir_missing: SecretsDirMissing | None = None,
        secrets_dir_max_size: int | None = None,
        secrets_case_sensitive: bool | None = None,
        secrets_prefix: str | None = None,
        secrets_nested_delimiter: str | None = None,
        secrets_nested_subdir: bool | None = None,
    ) -> None:
        # config options
        conf = settings_cls.model_config
        self.secrets_dir: str | None = first_not_none(
            secrets_dir,
            conf.get('secrets_dir'),
        )
        self.secrets_dir_missing: SecretsDirMissing | None = first_not_none(
            secrets_dir_missing,
            conf.get('secrets_dir_missing'),
            'warn',
        )
        self.secrets_dir_max_size: int = first_not_none(
            secrets_dir_max_size,
            conf.get('secrets_dir_max_size'),
            16 * 2**20,  # 8 MiB seems to be a reasonable default
        )
        self.case_sensitive: bool = first_not_none(
            secrets_case_sensitive,
            conf.get('secrets_case_sensitive'),
            conf.get('case_sensitive'),
            False,
        )
        self.secrets_prefix: str = first_not_none(
            secrets_prefix,
            conf.get('secrets_prefix'),
            conf.get('env_prefix'),
            '',
        )

        # nested options
        self.secrets_nested_delimiter: str | None = first_not_none(
            secrets_nested_delimiter,
            conf.get('secrets_nested_delimiter'),
            conf.get('env_nested_delimiter'),
        )
        self.secrets_nested_subdir: bool = first_not_none(
            secrets_nested_subdir,
            conf.get('secrets_nested_subdir'),
            False,
        )
        if self.secrets_nested_subdir:
            if secrets_nested_delimiter or conf.get('secrets_nested_delimiter'):
                raise SettingsError(
                    'Options secrets_nested_delimiter and secrets_nested_subdir '
                    'are mutually exclusive'
                )
            else:
                self.secrets_nested_delimiter = os.sep

        # ensure valid secrets_path
        if self.secrets_dir is None:
            self.secrets_path = None
        else:
            self.secrets_path: Path = Path(self.secrets_dir).expanduser().resolve()
            if not self.secrets_path.exists():
                match self.secrets_dir_missing:
                    case 'ok':
                        pass
                    case 'warn':
                        warnings.warn(f'directory "{self.secrets_path}" does not exist')
                    case 'error':
                        raise SettingsError(
                            f'directory "{self.secrets_path}" does not exist'
                        )
                    case _:
                        raise SettingsError(
                            f'invalid secrets_dir_missing value: '
                            f'{self.secrets_dir_missing}'
                        )
            else:
                if not self.secrets_path.is_dir():
                    raise SettingsError(
                        'secrets_dir must reference a directory, '
                        f'not a {path_type_label(self.secrets_path)}'
                    )
                secrets_dir_size = sum(
                    f.stat().st_size
                    for f in self.secrets_path.glob('**/*')
                    if f.is_file()
                )
                if secrets_dir_size > self.secrets_dir_max_size:
                    raise SettingsError(
                        f'secrets_dir size is above {self.secrets_dir_max_size} bytes'
                    )

        # construct parent
        super().__init__(
            settings_cls,
            case_sensitive=self.case_sensitive,
            env_prefix=self.secrets_prefix,
            env_nested_delimiter=self.secrets_nested_delimiter,
            env_ignore_empty=False,  # match SecretsSettingsSource behaviour
            env_parse_none_str=None,  # match SecretsSettingsSource behaviour
            env_parse_enums=True,  # match SecretsSettingsSource behaviour
        )
        self.env_parse_none_str = None  # update manually because of None

        # update parent members
        if self.secrets_path is None:
            self.env_vars = {}
        else:
            secrets = {
                str(p.relative_to(self.secrets_path)): p.read_text()
                for p in self.secrets_path.glob('**/*')
                if p.is_file()
            }
            self.env_vars = parse_env_vars(
                secrets,
                self.case_sensitive,
                self.env_ignore_empty,
                self.env_parse_none_str,
            )

    def __repr__(self) -> str:
        return f'FileSecretsSettingsSource(secrets_dir={self.secrets_dir!r})'

    def __call__(self):
        res = super().__call__()
        # breakpoint()  # this is the most informative place to debug
        return res


def first_not_none(*objs) -> Any:
    return next(filter(lambda o: o is not None, objs), None)



"""
These tests show that nested secrets problem exists.
"""

from pydantic import BaseModel
from pydantic_settings import BaseSettings, SettingsConfigDict
import pytest


def test_delimited_name_fails(monkeypatch, secrets_dir):
    class DbSettings(BaseModel):
        user: str
        password: str | None = None

    class Settings(BaseSettings):
        db: DbSettings
        app_key: str | None = None
        model_config = SettingsConfigDict(
            secrets_dir=secrets_dir,
            env_nested_delimiter='__',
        )

    monkeypatch.setenv('DB__USER', 'user')
    secrets_dir.add_files(
        ('app_key', 'secret'),
        ('db__password', 'secret'),  # file name with delimiter
    )

    conf = Settings()
    assert conf.app_key == 'secret'
    assert conf.db.password is None


def test_pure_name_fails(monkeypatch, secrets_dir):
    class DbSettings(BaseModel):
        user: str
        password: str | None = None

    class Settings(BaseSettings):
        db: DbSettings
        app_key: str | None = None
        model_config = SettingsConfigDict(
            secrets_dir=secrets_dir,
            env_nested_delimiter='__',
        )

    monkeypatch.setenv('DB__USER', 'user')
    secrets_dir.add_files(
        ('app_key', 'secret'),
        ('password', 'secret'),  # file name matching nested option name
    )

    conf = Settings()
    assert conf.app_key == 'secret'
    assert conf.db.password is None  # not loaded


def test_subdir_fails(monkeypatch, secrets_dir):
    class DbSettings(BaseModel):
        user: str
        password: str | None = None

    class Settings(BaseSettings):
        db: DbSettings
        app_key: str | None = None
        model_config = SettingsConfigDict(
            secrets_dir=secrets_dir,
            env_nested_delimiter='__',
        )

    monkeypatch.setenv('DB__USER', 'user')
    secrets_dir.add_files(
        ('app_key', 'secret'),
        ('db/password', 'secret'),  # file in nested subdirectory
    )

    with pytest.warns(UserWarning):
        conf = Settings()
    assert conf.app_key == 'secret'
    assert conf.db.password is None  # not loaded



"""
These tests show that the behaviour of FileSecretsSettingsSource matches
SecretsSettingsSource for options env_ignore_empty, env_parse_none_str, env_parse_enums.
"""

from enum import StrEnum

from pydantic_settings import BaseSettings, SettingsConfigDict
import pytest

from pydantic_file_secrets import FileSecretsSettingsSource


class SomeEnum(StrEnum):
    TEST = 'test'


class Settings(BaseSettings):
    key_empty: str | None = None
    key_none: str | None = None
    key_enum: SomeEnum | None = None


class SettingsPairMaker:
    def __call__(
        self,
        model_config: SettingsConfigDict,
    ) -> tuple[type[Settings], type[Settings]]:
        class SettingsSSS(Settings):  # SecretsSettingsSource
            pass

        class SettingsFSSS(Settings):  # FileSecretsSettingsSource
            @classmethod
            def settings_customise_sources(
                cls,
                settings_cls,
                init_settings,
                env_settings,
                dotenv_settings,
                file_secret_settings,
            ) -> tuple:
                return (
                    env_settings,
                    init_settings,
                    FileSecretsSettingsSource(settings_cls),
                )

        SettingsSSS.model_config = model_config
        SettingsFSSS.model_config = model_config

        return (SettingsSSS, SettingsFSSS)


@pytest.fixture()
def settings_models() -> SettingsPairMaker:
    return SettingsPairMaker()


@pytest.fixture()
def populated_secrets_dir(secrets_dir):
    secrets_dir.add_files(
        ('key_empty', ''),
        ('key_none', 'null'),
        ('key_enum', 'test'),
    )
    yield secrets_dir


@pytest.mark.parametrize(
    'conf', [{}, {'env_ignore_empty': True}, {'env_ignore_empty': False}]
)
def test_env_ignore_empty(settings_models, conf, populated_secrets_dir):
    SettingsSSS, SettingsFSSS = settings_models(
        model_config=SettingsConfigDict(
            secrets_dir=populated_secrets_dir,
            **conf,
        ),
    )
    conf_sss, conf_fsss = SettingsSSS(), SettingsFSSS()
    assert conf_sss.key_empty == conf_fsss.key_empty == ''


@pytest.mark.parametrize('conf', [{}, {'env_parse_none_str': 'null'}])
def test_env_parse_none_str(settings_models, conf, populated_secrets_dir):
    SettingsSSS, SettingsFSSS = settings_models(
        model_config=SettingsConfigDict(
            secrets_dir=populated_secrets_dir,
            **conf,
        ),
    )
    conf_sss, conf_fsss = SettingsSSS(), SettingsFSSS()
    assert conf_sss.key_none == conf_fsss.key_none == 'null'


@pytest.mark.parametrize(
    'conf', [{}, {'env_parse_enums': True}, {'env_parse_enums': False}]
)
def test_env_parse_enums(settings_models, conf, populated_secrets_dir):
    SettingsSSS, SettingsFSSS = settings_models(
        model_config=SettingsConfigDict(
            secrets_dir=populated_secrets_dir,
            **conf,
        ),
    )
    conf_sss, conf_fsss = SettingsSSS(), SettingsFSSS()
    assert conf_sss.key_enum is SomeEnum.TEST
    assert conf_fsss.key_enum is SomeEnum.TEST



def test_delimited_name(settings_model, monkeypatch, secrets_dir):
    monkeypatch.setenv('DB__USER', 'user')
    secrets_dir.add_files(
        ('app_key', 'secret1'),
        ('db___password', 'secret2'),  # file name with delimiter
    )
    Settings = settings_model(
        model_config=dict(
            env_nested_delimiter='__',
            secrets_dir=secrets_dir,
            secrets_nested_delimiter='___',
        ),
    )
    conf = Settings()
    assert conf.app_key == 'secret1'
    assert conf.db.password == 'secret2'  # noqa: S105


def test_secrets_off(settings_model, monkeypatch, secrets_dir):
    monkeypatch.setenv('DB__USER', 'user')
    secrets_dir.add_files(
        ('app_key', 'secret1'),
        ('db__password', 'secret2'),  # file name with delimiter
    )
    Settings = settings_model(
        model_config=dict(
            env_nested_delimiter='__',
            # missing secrets_dir
        ),
    )
    conf = Settings()
    assert conf.app_key is None
    assert conf.db.password is None
